#!/usr/bin/perl
#
# $Id: paraloop.pl 73 2005-06-23 09:43:42Z manu $
#

=head1 Copyright Notice

 Copyright I.N.R.A. - C.N.R.S. 

 emmanuel.courcelle@toulouse.inra.fr
 jerome.gouzy@toulouse.inra.fr

 This software is a computer program whose purpose is to help you run 
 your scripts on parallel computers.

 This software is governed by the CeCILL  license under French law and
 abiding by the rules of distribution of free software.  You can  use, 
 modify and/ or redistribute the software under the terms of the CeCILL
 license as circulated by CEA, CNRS and INRIA at the following URL
 http://www.cecill.info" 

 As a counterpart to the access to the source code and  rights to copy,
 modify and redistribute granted by the license, users are provided only
 with a limited warranty  and the software's author,  the holder of the
 economic rights,  and the successive licensors  have only  limited
 liability. 

 In this respect, the user's attention is drawn to the risks associated
 with loading,  using,  modifying and/or developing or reproducing the
 software by the user in light of its specific status of free software,
 that may mean  that it is complicated to manipulate,  and  that  also
 therefore means  that it is reserved for developers  and  experienced
 professionals having in-depth computer knowledge. Users are therefore
 encouraged to load and test the software's suitability as regards their
 requirements in conditions enabling the security of their systems and/or 
 data to be ensured and,  more generally, to use and operate it in the 
 same conditions as regards security. 

 The fact that you are presently reading this means that you have had
 knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

paraloop.pl - Run something in a PARAllel LOOP

=head1 SYNOPSIS

paraloop.pl [--switches | --parameters | --plugins]

   --switches    ==> to know everything about available switches
   --parameters  ==> to know everything about parameters
   --plugins     ==> to know the list of installed plugins
   --schedulers  ==> to know the list of installed schedulers

=head1 DESCRIPTION

Execute a loop in parallel on several processors or several nodes of a cluster.

=head1 SUBROUTINES

=cut

################################################################################
#
# THE PACKAGE MAIN
#
################################################################################

package main;
use strict;
use warnings;

use FileHandle;
use Getopt::Long;
use IO::File;
use Cwd;

BEGIN
{
    if (exists $ENV{'PARALOOP'})
    {
        unshift(@INC, $ENV{'PARALOOP'}.'/lib');
    }
    elsif (exists $ENV{'LIPMLIB'})
    {
        unshift(@INC, $ENV{'LIPMLIB'});
    }
    else
    {
        my ($dirprg, $nameprg) = $0 =~ /(.+)\/(.+)/;
        unshift(@INC, "$dirprg/lib", "$dirprg/../lib");
    }
}

use ParamParser;
use Logger;
use Runner;

#
# some const
#

use constant MAXFILESIZE       => 1000;
use constant NAMESPACE         => 'PARALOOP_';
use constant LOCK_DIR          => NAMESPACE . 'lock';
use constant ERROR_DIR         => NAMESPACE . 'error';
use constant LOG_LEVEL_DEFAULT => '01';
use constant WAIT_TIME         => 100;
#use constant WAIT_TIME         => 3;

#
# M A I N   P R O G R A M
#

my $VERSION = do {my @r = (q$Rev: 73 $ =~ /\d+/g); $r[0]};

#
# Was paraloop.pl ran by itself ? we test the existence of the environment variable subcount
# This variable is SET INTERNALLY BY THIS SCRIPT - DO NOT SET THE VARIABLE MANUALLY !!!
#

my $subcount_variable = NAMESPACE . 'subcount';
if (!exists $ENV{$subcount_variable})
{
    &MainFromCommandLineThenExit();    # Was called by user
}
else
{
    &MainFromSubmitThenExit();         # Was called by paraloop through the Submit() procedure.
}

#
# Title: MainFromCommandLineThenExit
# Usage: &MainFromCommandLineThenExit()
# procedure: Called when paraloop is entered through a command line
#            The caller may be a human, or another instance of paraloop
# Args:      None
# Return:    NEVER RETURN
# Globals:   none
#

sub MainFromCommandLineThenExit
{
    my $o_params = New ParamParser();    # The object handling parameters
    $o_params->SetUsage(\&Usage);
    $o_params->SetBehaviour('exit_on_getopt_error');
    $o_params->SelectNameSpace(NAMESPACE);
    $o_params->Update(
                      'GETOPTLONG', 'I',              "restart=s",    "cfg=s",
                      "db=s",       "queue=s",        "local",        "output=s",
                      "input=s",    "slice_offset=i", "slice_size=i", "slice_nb=i",
                      "start=i",    "end=i",          "step=i",       "program=s",
                      "account=s",  "ncpus=i",        "monoprocessor","interleaved",  "verbose",
                      "quiet",      "help",           "switches",     "parameters", "waitonly=s",
                      "plugins",    "schedulers",     "wait",         "plugin_debug"
                      );

    unless ($o_params->IsDefined('quiet'))
    {
        print "paraloop.pl - VERSION $VERSION - Using ParamParser v$ParamParser::VERSION, Logger v$Logger::VERSION\n";
    }
    
    my $plugin_debug = ($o_params->IsDefined('plugin_debug')) ? 1 : 0;
    &Switches($plugin_debug)   if ($o_params->IsDefined('switches'));
    &Parameters($plugin_debug) if ($o_params->IsDefined('parameters'));
    &Plugins($plugin_debug)    if ($o_params->IsDefined('plugins'));
    &Schedulers($plugin_debug) if ($o_params->IsDefined('schedulers'));

    my $verbose = ($o_params->IsDefined('verbose')) ? 1 : 0;
    my $quiet   = ($o_params->IsDefined('quiet'))   ? 1 : 0;
    my $monoprocessor = ($o_params->IsDefined('monoprocessor')) ? 1 : 0;
    if ($verbose == 1 and $quiet == 1)
    {
        die "\nERROR - do NOT specify --verbose together with --quiet\n";
        #&Usage();
    }

    #
    # monoprocessor specified: start one job, with (start,end,step) specified or with (slice_offset,slice_size,slice_nb) specified
    #
    
    if ($monoprocessor==1)
    {
	MainCommandLineOneLoopThenExit($o_params);
    }
    
    #
    # restart specified: restart from an interrupted job, or from a monoprocessor interrupted job
    #
    
    elsif ($o_params->IsDefined('restart'))
    {
        MainCommandLineRestartThenExit($o_params);
    }
    
    #
    # waitonly specified: wait for the end of all jobs, or for the end of a monoprocessor job
    #
    
    elsif ($o_params->IsDefined('waitonly'))
    {
        MainCommandLineWaitOnlyThenExit($o_params);
    }
    
    #
    # ncpus mode: call paraloop from a high level loop (ncpus iterations)
    #
    
    else
    {
	MainCommandLineNcpusThenExit($o_params);
    }
}

#
# Title: MainFromCommandLineNcpusThenExit
# Usage: &MainFromCommandLineNcpusThenExit($o_params)
# procedure: Called by MainFromCommandLineThenExit when paraloop is entered
#            through a command line with the ncpus switch
# Args:      $o_params
# Return:    NEVER RETURN
# Globals:   none
#

# MODE = multi/init

sub MainCommandLineNcpusThenExit
{
    
    my $o_params = shift;
    #
    # Update $o_params ONLY from command line: Only some parameters allowed
    #

    $o_params->Init();               # Workaround a bug in ParamParser
    $o_params->SetUsage(\&Usage);    # kifkif
    $o_params->SelectNameSpace();    # NO Namespace selected, because used ONLY for the command line of children

    #$o_params->SetUsage(\&Usage);
    $o_params->SetBehaviour('exit_on_getopt_error');
    $o_params->Update(
                      'GETOPTLONG', 'I',           "cfg=s",   "db=s",    "queue=s",   "local",
                      "output=s",   "input=s",     "start=i", "end=i",   "program=s", "account=s",
                      "ncpus=i",    "interleaved", "wait",    "verbose", "quiet",     "plugin_debug",
                      "help"
                      );

    #
    # Create an alternate parameters object, with data coming ALSO from the configuration files.
    # The parameters inside $o_alt_params will NOT be propagated to paraloop through command line
    # N.B. It would be better here to have a clone() function in ParamParser
    #

    my $o_alt_params = New ParamParser();
    $o_alt_params->SelectNameSpace(NAMESPACE);
    $o_alt_params->SetUsage(\&Usage);
    $o_alt_params->Update(
                      'GETOPTLONG', 'I',           "cfg=s",   "db=s",    "queue=s",   "local",
                      "output=s",   "input=s",     "start=i", "end=i",   "program=s", "account=s",
                      "ncpus=i",    "interleaved", "wait",    "verbose", "quiet",     "plugin_debug",
                      "help"
                      );
    
    $o_alt_params->Set('cfg', $o_params->Get('cfg')) if ($o_params->IsDefined('cfg'));
    &ReadCfgFiles($o_alt_params);    # Useful for reading the cluster_nodes, ncpus, or other paramters
    
    $o_params->AssertDefined('program','input','output');

    # Init some local variables from those parameters

    my $verbose = ($o_params->IsDefined('verbose')) ? 1 : 0;
    my $quiet   = ($o_params->IsDefined('quiet'))   ? 1 : 0;

    my $lock_directory;
    if ($o_alt_params->IsDefined('lock_directory'))
    {
        $lock_directory = $o_alt_params->Get('lock_directory');
    }
    else
    {
        $lock_directory = LOCK_DIR;
    }

    # Create a log object, to log to $output.log (WITHOUT the numeric fields)
    
    my $output = $o_alt_params->Get("output");
    my $log_file .= $output . '.log';
    $o_alt_params->Set('log_file', $log_file);
    my $log_level;
    if ($o_alt_params->IsDefined("log_level"))
    {
        $log_level = $o_alt_params->Get("log_level");
    }
    else
    {
        $log_level = LOG_LEVEL_DEFAULT;
        $o_alt_params->Set("log_level", $log_level);
    }

    my $o_log = Logger->New(">>$log_file", $log_level);

    # Init the $ncpus variable:
    #      If the ncpus switch was specified, we take its value.
    #      Else, we look for the parameter PARALOOP_<Scheduler>_ncpus, where <Scheduler> is the name of the chosen scheduler.
    #      Else, we die with an error message (NO default value !)
    #
    
    my $ncpus='';
    my $scheduler='';
    if ($o_alt_params->IsDefined('local'))
    {
        $scheduler = 'System';
        $o_log->Trace("Scheduler $scheduler selected (--local)\n",".","1");
    }
    elsif ($o_alt_params->IsDefined('Scheduler')) 
    {
        $scheduler = $o_alt_params->Get('Scheduler');
        $o_log->Trace("Scheduler $scheduler selected\n",".","1");
    }
    else
    {
        $o_log->Trace("ERROR - The Scheduler parameter should be defined\n",".","1");
        die "\nERROR - The Scheduler parameter should be defined\n";
    };
    
    if ($o_alt_params->IsDefined('ncpus'))
    {
        $ncpus = $o_alt_params->Get('ncpus');
    }
    else
    {
        my $def_ncpus_param = $scheduler . '_ncpus';
        if ($o_alt_params->IsDefined($def_ncpus_param))
        {
            $ncpus = $o_alt_params->Get($def_ncpus_param);
        }
        else
        {
            $o_log->Trace("ERROR - The ncpus switch OR the $def_ncpus_param parameter should be defined\n",".","1");
            die "\nERROR - The ncpus switch OR the PARALOOP_$def_ncpus_param parameter should be defined\n";
        }
    }
    $o_log->Trace("ncpus = $ncpus\n",".","1");
   
    # The plugin is created to check everything is OK, to print a welcome message and to know the number of records

    my $plugin = &CreatePlugin($o_alt_params,$o_log);

    # Set the quiet options for the children, avoiding a lot of redundant messages
    $o_params->Delete('verbose');
    $o_params->Set('quiet', 1);

    # Delete the wait switch for the children
    my $wait = ($o_params->IsDefined('wait')) ? 1 : 0;
    $o_params->Delete('wait');

    # Delete the ncpus switch for the children
    #my $ncpus = $o_params->Get('ncpus');
    #$o_params->Delete('ncpus');

    # Set the monoprocessor switch for the children
    $o_params->Set('monoprocessor',1);
    
    # Compute start,end values (may be nothing specified here)
    my ($start, $end);
    if ($o_params->IsDefined('start'))
    {
        $start = $o_params->Get('start');
    }
    else
    {
        $start = 0;
    }
    print "start = $start\n" unless ($quiet == 1);

    if ($o_params->IsDefined('end'))
    {
        $end = $o_params->Get('end');
    }
    else
    {
        print "Counting the number of records in the input files\n" unless ($quiet == 1);
        my $nb_seq = $plugin->TellLength();
        $end = $nb_seq - 1;
        print "Found $nb_seq records.\n" if ($verbose == 1);
	$o_log->Trace("Found $nb_seq records.\n",".","1");
    }
    print "end = $end\n" unless ($quiet == 1);
    
    if ($end-$start < $ncpus) {		# Could happen, specially with the Shell plugin
        my $old_ncpus = $ncpus; 
        $ncpus = $end-$start + 1;
        print "WARNING - ncpus was changed from $old_ncpus to $ncpus\n";
        $o_log->Trace("WARNING - ncpus was changed from $old_ncpus to $ncpus.\n",".","1");
    };
    
    # The object is not needed anymore

    undef($plugin);

    if ($o_params->IsDefined('plugin_debug'))
    {
        &PrintVar($o_params);
        print "Now leaving - Please restart without the plugin_debug switch\n";
        exit(0);
    }

    #
    # Interleaved mode: the load is distributed as follows (for ncpus=4):
    #                   CPU0 => 0,4,8,...
    #                   CPU1 => 1,5,9,...
    #                   CPU2 => 2,6,10,...
    #                   CPU3 => 3,7,11,...
    # WARNING: If the end specified was 10, the REAL LAST read record will be 11
    #

    if ($o_params->IsDefined('interleaved'))
    {
        $o_params->Delete('interleaved');
        $o_params->Set('end', $end);
        $o_params->Set('step', ($end > $start) ? $ncpus : -$ncpus);
        for (my $cpu = 0; $cpu < $ncpus; $cpu++)
        {
            $o_params->Set('start', ($end > $start) ? $start + $cpu : $start - $cpu);
            $o_params->Dump('GETOPTLONG');
            my $cmd = $0 . ' ' . join(' ', @ARGV);
            print "calling $cmd\n" if ($verbose == 1);
	    $o_log->Trace("INTERLEAVED MODE - STARTING PARALOOP FOR CPU $cpu\n",'.','1');
            &RunExtNoWait(-cmd => $cmd);
        }
        &WaitParaloop($lock_directory) if ($wait == 1);
        exit 0;
    }

    #
    # Normal mode: the input file is splitted in splices
    #

    else
    {
        my $total_nb = $end - $start + 1;
        my $slice_size;
        my $slice_offset;
        if ($total_nb % $ncpus == 0)
        {
            $slice_size = $total_nb / $ncpus;
        }
        else
        {
            $slice_size = int($total_nb / $ncpus) + 1;
        }
        $o_params->Delete('start');
        $o_params->Delete('end');
        $o_params->Set('slice_offset', $start);
        $o_params->Set('slice_size',   $slice_size);
        for (my $cpu = 0; $cpu < $ncpus; $cpu++)
        {
            $o_params->Set('slice_nb', $cpu);
            $o_params->Dump('GETOPTLONG');
            my $cmd = $0 . ' ' . join(' ', @ARGV);
            print "calling $cmd\n" if ($verbose == 1);
 	    $o_log->Trace("NORMAL MODE - STARTING PARALOOP FOR SLICE $cpu\n",'.','1');
           &RunExtNoWait(-cmd => $cmd);
        }
        &WaitParaloop($lock_directory) if ($wait == 1);
        exit 0;
    }
}

#
# Title: MainFromCommandLineRestartThenExit
# Usage: &MainFromCommandLineRestartThenExit($o_params)
# procedure: Called by MainFromCommandLineThenExit when paraloop is entered
#            through a command line with the restart switch
# Args:      $o_params
# Return:    NEVER RETURN
# Globals:   None
#

sub MainCommandLineRestartThenExit
{
    my $o_params = shift;
    die "\nplugin_debug cannot be specified with --restart\n" if ($o_params->IsDefined('plugin_debug'));

    my $restart_file = $o_params->Get('restart');
    if (!-e $restart_file)
    {
        die "ERROR - The restart file $restart_file should exist";
    }

    #
    # Directory specified: look for every lock file here, and call paraloop with the correct switch
    #

    if (-d $restart_file)
    {
        MainCommandLineRestartIsDirThenExit($o_params,$restart_file);
    }
    else
    {
        MainCommandLineRestartIsFileThenExit($o_params,$restart_file);
    }
}

#
# Title: MainFromCommandLineRestartIsDirThenExit
# Usage: &MainFromCommandLineRestartIsDirThenExit($o_params,$restart_dir)
# procedure: Called by MainFromCommandLineRestartThenExit when paraloop is entered
#            through a command line with the restart switch AND the argument of the
#            switch is a directory
# Args:      $o_params    = The parameters
#            $restart_dir = The lock directory
# Return:    NEVER RETURN
# Globals:   None
#

# MODE = multi/restart

sub MainCommandLineRestartIsDirThenExit
{
    my $o_params = shift;
    my $restart_dir = shift;
    my $wait = ($o_params->IsDefined('wait')) ? 1 : 0;
    $o_params->Delete('wait');

    foreach my $f (glob("$restart_dir/lock.*"))
    {    # For each lock file...
        next if (-d $f);    # Should not happen !!!
        my $cmd = "$0 --restart $f";
        print "calling $cmd\n" if ($o_params->IsDefined('verbose'));
        &RunExtNoWait(-cmd => $cmd);
    }
    &WaitParaloop($restart_dir) if ($wait == 1);
    exit 0;
}

#
# Title: MainFromCommandLineRestartIsFileThenExit
# Usage: &MainFromCommandLineRestartIsfileThenExit($o_params,$restart_file)
# procedure: Called by MainFromCommandLineRestartThenExit when paraloop is entered
#            through a command line with the restart switch AND the argument of the
#            switch is a lock file
# Args:      $o_params = The parameters object      
#$restart_file = The lock file to restart from
# Return:    NEVER RETURN
# Globals:   None
#

# MODE = mono/restart

sub MainCommandLineRestartIsFileThenExit
{
    my $o_params = shift;
    my $restart_file = shift;

    #
    # File specified: restart paraloop reading the parameters from this file
    #

    $o_params->SelectNameSpace();    # Cancel the namespace before reading the files
    $o_params->Update($restart_file, 'O');    # Read the parameters from the file that was saved then
    $o_params->SelectNameSpace(NAMESPACE);    # The conf files are NOT used
    my $lock_file = $o_params->Get('lock_file');
    if (!-e $lock_file)
    {
        die "ERROR - The lock file $lock_file should exist.";
    }

    unlink($o_params->Get('stop_all'), $o_params->Get('stop_this'));    # Remove stop files, if any
    my $job_id = $o_params->Get('job_id');
    $o_params->Delete('subcount');    # Delete subcount, because we are restarted by the user

    my $log_file  = $o_params->Get('log_file');
    my $log_level = $o_params->Get("log_level");
    my $o_log = Logger->New(">>$log_file", $log_level);

    my $sched = &CreateScheduler($o_params,$o_log);
    my $state = $sched->IsLaunched($job_id);
    if ($state != 0)
    {
        die "ERROR - The job $job_id (see $lock_file) is already launched";
    }

    $o_log->Trace("RESTARTING PARALOOP FROM THE FILE $restart_file\n", '*', '1');
    $o_log->Trace("PARAMETERS WERE SET FROM THE FILE $lock_file\n",    '.', '2');

    $o_params->Set('status', 'RESTARTING');

    $o_params->SelectNameSpace();
    $o_params->Dump('ENV', NAMESPACE);    # Copy the parameters to ENV, with a prefix for identification.
    $o_params->SelectNameSpace(NAMESPACE);

    $job_id = $sched->Submit();
    print "launched job, job ID = $job_id\n";
    exit(0);
}

#
# Title: MainFromCommandLineWaitOnlyThenExit
# Usage: &MainFromCommandLineWaitOnlyThenExit($o_params)
# procedure: Called by MainFromCommandLineThenExit when paraloop is entered
#            through a command line with the waitonly switch
# Args:      $o_params
# Return:    NEVER RETURN
# Globals:   None
#

# MODE = waiting


sub MainCommandLineWaitOnlyThenExit
{
    my $o_params = shift;
    my $waitonly_file = $o_params->Get('waitonly');
    &WaitParaloop($waitonly_file);
}

#
# Title: MainFromCommandLineOneLoopThenExit
# Usage: &MainFromCommandLineOneLoopThenExit($o_params)
# procedure: Called by MainFromCommandLineThenExit when paraloop is entered
#            through a command line for executing a monoprocessor loop
# Args:      $o_params
# Return:    NEVER RETURN
# Globals:   None
#

# MODE = mono/init

sub MainCommandLineOneLoopThenExit
{    # STARTING A NEW JOB

    my $o_params = shift;
    my $ra_cfg_files = &ReadCfgFiles($o_params);
    $o_params->AssertDefined("output", "input", "program");

    $o_params->Set('workdir', cwd);    # Keep track of the current directory for the qsub calls

    my ($db, $start, $end, $slice_offset, $slice_size, $slice_nb) = ('', -1, -1, 0, -1, -1);
    $db           = $o_params->Get("db")               if ($o_params->IsDefined("db"));
    $start        = $o_params->Get("start") + 0        if ($o_params->IsDefined("start"));
    $end          = $o_params->Get("end") + 0          if ($o_params->IsDefined("end"));
    $slice_offset = $o_params->Get("slice_offset") + 0 if ($o_params->IsDefined("slice_offset"));
    $slice_size   = $o_params->Get("slice_size") + 0   if ($o_params->IsDefined("slice_size"));
    $slice_nb     = $o_params->Get("slice_nb") + 0     if ($o_params->IsDefined("slice_nb"));

    if ($start == -1 or $end == -1)
    {
        if ($o_params->IsDefined('step'))
	{
	    die "ERROR - The switch --step must be used with the switches --start and --end\n";
	};
        if ($slice_size == 0 or $slice_nb == -1)
	{
	    die "ERROR - Please specify --start,--end OR --slice_size,--slice_nb\n";
	};
        $start = $slice_offset + $slice_size * $slice_nb;
        $end   = $start + $slice_size - 1;                  # The job is done in the interval [$start,$end]
        $o_params->Set('start',   $start);   # start has the same value across the many qsubs or interruptions
        $o_params->Set('current', $start);   # current is updated when the job resubmit itself or when it is interrupted
        $o_params->Set('end',     $end);
    }
    my $step;
    if ($o_params->IsDefined('step') == 0)
    {
        $step = 1;
        $o_params->Set('step', 1);
    }
    else
    {
        $step = $o_params->Get('step');
    }

    die "step must be >0 or <0"                 if ($step == 0);
    die "start and end must be > 0"             if ($start < 0 or $end < 0);
    die "end must be > start, because step > 0" if ($end < $start and $step > 0);
    die "end must be < start, because step < 0" if ($end > $start and $step < 0);

    my $current = $start;
    $o_params->Set('current', $current);

    my $output = $o_params->Get("output");
    my $log_file .= $output . '.' . $start . '.0.log';
    $o_params->Set('log_file', $log_file);

    $output .= '.' . $start . '.0.txt';
    $o_params->Set('output', $output);

    my $log_level;
    if ($o_params->IsDefined("log_level"))
    {
        $log_level = $o_params->Get("log_level");
    }
    else
    {
        $log_level = LOG_LEVEL_DEFAULT;
        $o_params->Set("log_level", $log_level);
    }
    my $o_log = Logger->New(">$log_file", $log_level);

    my $plugin = &CreatePlugin($o_params,$o_log);
    if ($step < 0)
    {
        $plugin->SkipRecords(-1);    # Just to check that this object is able to skip a negative number of records
    }

    my $max_file_size;
    if ($o_params->IsDefined('max_file_size'))
    {
        $max_file_size = $o_params->Get('max_file_size');
    }
    else
    {
        $max_file_size = MAXFILESIZE;
        $o_params->Set('max_file_size', $max_file_size);
    }

    my $lock_directory;
    if ($o_params->IsDefined('lock_directory'))
    {
        $lock_directory = $o_params->Get('lock_directory');
    }
    else
    {
        $lock_directory = LOCK_DIR;
    }
    unless (-d $lock_directory)
    {
        mkdir $lock_directory;
    }

    my $error_directory;
    if ($o_params->IsDefined('error_directory'))
    {
        $error_directory = $o_params->Get('error_directory');
    }
    else
    {
        $error_directory = ERROR_DIR;
    }

    $o_params->Set('error_directory', $error_directory);
    unless (-d $error_directory)
    {
        mkdir $error_directory;
    }

    my $sched = &CreateScheduler($o_params,$o_log);

    my $lock_file = $lock_directory . "/lock.$start.lck";
    $o_params->Set('lock_file', $lock_file);
    if (-e $lock_file)
    {
        die "ERROR - LOCKFILE $lock_file FOUND";
    }
    my $lck = new IO::File(">$lock_file") or die "Cannot create $lock_file";
    print "created $lock_file\n";

    my $stop_all  = $lock_directory . '/paraloop.stp';
    my $stop_this = $lock_directory . "/paraloop.$start.stp";
    $o_params->Set('stop_all',  $stop_all);
    $o_params->Set('stop_this', $stop_this);

    $o_log->Trace("STARTING PARALOOP\n",                         '*', '1');
    $o_log->Trace("PARAMETERS WERE SET FROM THE COMMAND LINE\n", '.', '2');
    for my $f (@$ra_cfg_files)
    {
        $o_log->Trace("PARAMETERS WERE SET FROM THE FILE $f\n", '.', '2');
    }

    $o_params->Set('status', 'STARTING');

    $o_params->Dump('ENV', NAMESPACE);    # Copy the parameters to ENV, with a prefix for identification.

    #
    # debug mode: Start MainLoop without forking a new process:
    #             This is convenient only for debugging a new pligin
    #

    # NOTE - We are now leaving the mono/init mode: in the normal operation we submit
    #        a new paraloop job, but in debug mode we just switch to mono/running and we call MainLoop
    
    if ($o_params->IsDefined('plugin_debug'))
    {                                     # Start MainLoop without forking a new process
        &MainLoop($o_params,$o_log,$plugin,$sched);
        &PrintVar($o_params);
    }
    else
    {
        my $job_id = $sched->Submit();
        print "launched job, job ID = $job_id\n";
    }
    exit(0);
}

#
# Title: MainFromSubmitThenExit
# Usage: &MainFromSubmitThenExit()
# procedure: Called when paraloop is entered after a call to the Submit sub
#            of a previous instance
# Args:      None
# Return:    NEVER RETURN
# Globals:   None
#

# MODE = mono/running

sub MainFromSubmitThenExit
{
    my $o_params = New ParamParser();    # The object handling parameters
    $o_params->SetUsage(\&Usage);
    $o_params->SetBehaviour('exit_on_getopt_error');
    $o_params->Update('ENV', 'I');    # Read parameters from ENV. Some of them start with a prefix
    $o_params->SelectNameSpace(NAMESPACE);    # Consider this prefix as a space name
    my $lock_file = $o_params->Get('lock_file');    # Read last parameters (namely job_id) from the lock file
                                                    #&PrintVar($o_params);

    chdir($o_params->Get('workdir'));

    my $log_file  = $o_params->Get('log_file');
    my $log_level = $o_params->Get("log_level");
    my $o_log = Logger->New(">>$log_file", $log_level);
    $o_log->Trace("PARAMETERS WERE SET FROM ENV\n", '.', '2');

    # Create the Scheduler object
    my $sched = &CreateScheduler($o_params,$o_log);

    # Ask the scheduler for the job Id, and store this ID in the parameters
    my $job_id = $sched->JobId();
    $o_params->Set('job_id', $job_id);

    # Create the Plugin object
    my $plugin = &CreatePlugin($o_params,$o_log);

    &MainLoop($o_params,$o_log,$plugin,$sched);

    # We arrive here ONLY if the main loop is terminated

    unlink($lock_file);    # remove the lock file
    my $lock_directory = $o_params->Get('lock_directory');
    rmdir($lock_directory);    # try to remove the lock directory: should work if it is the LAST job
    exit(0);

}

#
# Title: MainLoop
# Usage: &MainLoop($o_params,$o_log,$o_plugin,$o_scheduler)
# procedure: Enter the main loop of the program
#            1/ Retrieve 4 objects: params, log, plugin, scheduler
#            2/ Skip the beginning of the file if necessary, reaching the correct start record
#            3/ Loop from current to end. At each iteration:
#                     call $plugin->Exec(), doing what must be done
#                     If the output file is too big, change output and log file
#                     If time spent in the queue is too high, call qsub and never return
#                     If SIGINT was received, OR if p.stp was created, OR if paraloop.START.stp was created,
#                     call StopJob and never return
#            4/ return only if END was reached
#
# Return:    Nothing
# Args:      4 objects
# Globals:   none
#

sub MainLoop
{
    my $o_params        = shift;
    my $o_log           = shift;
    my $o_plugin        = shift;
    my $o_sched         = shift;
    my $lock_file       = $o_params->Get('lock_file');
    my $start           = $o_params->Get('start') + 0;
    my $current         = $o_params->Get('current') + 0;
    my $step            = $o_params->Get('step') + 0;
    my $end             = $o_params->Get('end') + 0;
    my $fair_time_limit = 0 + $o_params->IsDefined('fair_time_limit') ? $o_params->Get('fair_time_limit') : 0;
    my $db              = '';
    $db = $o_params->Get('db') if ($o_params->IsDefined("db"));
    my $input       = $o_params->Get('input');
    my $output      = $o_params->Get('output');
    my $output_size = (-s $output);
    my $program     = $o_params->Get('program');

    #
    # skip the first n sequences
    #

    if ($current >= 0)
    {    # Nothing to do if $current < 0

        $o_params->Set('status', 'RUNNING');
        $o_params->Dump($lock_file, NAMESPACE);    # update the lock file

        if ($current > 0)
        {
            my $last_skipped = $current - 1;
            $o_log->Trace("SKIPPING SEQUENCES FROM 0 TO $last_skipped\n", '*', '1');
            $o_plugin->SkipRecords($current);
        }
        my $i_end           = 1 + ($end - $current) / $step;
        my $first_iteration = 1;
        my $start_time      = time();
        my $elapsed_time    = 0;
        for (my $i = 0; $i < $i_end; $i++)
        {

            last if ($step>0 and $o_plugin->Tell()>$end);
            last if ($step<0 and $o_plugin->Tell()<$end);


            #
            # Do we still have time ?
            #

            my $timeout = $o_sched->RemTime();
            if ($timeout == 0)
            {
                $o_params->Dump('ENV');    # Copy the parameters to ENV, without adding any prefix
                $o_params->Set('status', 'SUBMITTING A NEW JOB');
                $o_sched->Submit();
                exit;                      # DO NOT RETURN FROM MAINLOOP
            }
            if ($timeout > 0)
            {
                $o_log->Trace("THIS JOB WILL BE INTERRUPTED BY THE SCHEDULER IN $timeout s\n", '.', '2');
            }
            if ($fair_time_limit != 0)
            {
                my $fair_remaining_time = $fair_time_limit - $elapsed_time;
                $o_log->Trace("fair_time_limit=$fair_time_limit s; TIME REMAINING=$fair_remaining_time s\n", '.', '2');
            }

            my ($incr_size, $sts, $kld) = $o_plugin->Exec($timeout);    # EXECUTE THE PROG
            ############################### my ($incr_size,$sts,$kld) = $o_plugin->Exec(4);	# EXECUTE THE PROG
            $sts = 0 if (!defined($sts));                             # Some objects do not care of return status
            $kld = 0 if (!defined($kld));                             # Some objects do not care of timeout
            if ($sts != 0)
            {
                $o_log->Trace("THE EXEC FUNCTION RETURNED $sts\n", ".", '2');
                last;
            }
            if (($kld == 0) or ($kld == -1 and $first_iteration == 1))
            {    # If all OK, or if timeout detected at 1st iteration:
                if ($kld == -1)
                {    # start next step in either case
                    $o_log->Trace("THE COMMAND TAKES TOO MUCH TIME FOR THIS QUEUE\n", ".", '1');
                }

                #
                # Keep the lock file synchronized with the program
                #

                $o_plugin->SkipRecords($step - 1) if ($step != 1);    # skip some records, unless step=1
                $current += $step;                                  # current is the NEXT record number
                $o_params->Set('current', $current);
                $o_params->Set('status',  'RUNNING');
                $first_iteration = 0;
            }

            if ($kld == -1)
            {                                                       # Timeout reached: resubmit
                $o_log->Trace("NO MORE TIME IN QUEUE\n", ".", '1');
                $o_params->Dump('ENV');    # Copy the paramters to ENV, without adding any prefix
                $o_params->Set('status', 'SUBMITTING A NEW JOB');
                $o_sched->Submit();
                exit;                      # DO NOT RETURN FROM MAINLOOP
            }

            $elapsed_time = time() - $start_time;
            if ($fair_time_limit > 0 and $elapsed_time >= $fair_time_limit)
            {
                $o_log->Trace(
                        "TOO MUCH TIME SPENT IN THIS QUEUE (elapsed $elapsed_time, fair_time_limit $fair_time_limit)\n",
                        ".", '1');
                $o_params->Dump('ENV');    # Copy the parameters to ENV, without adding any prefix
                $o_params->Set('status', 'SUBMITTING A NEW JOB');
                $o_sched->Submit();
                exit;                      # DO NOT RETURN FROM MAINLOOP
            }

            $o_params->Dump($lock_file, NAMESPACE);    # update the lock file

            #
            # Size matters: we don't want the output file to become too large
            #

            $output_size += $incr_size;
            my $max_file_size = $o_params->Get('max_file_size') + 0;
            if ($output_size / 1000000 >= $max_file_size)
            {
                my $log_file      = $o_params->Get('log_file');
                my $next_log_file = &IncrementName($log_file);
                my $next_output   = &IncrementName($output);
                $o_log->Trace("NOW WRITING OUTPUT TO $next_output\n", ".", '1');
                $o_params->Set('log_file', $next_log_file);
                $o_log->SetLogFile($next_log_file);
                $o_params->Set('output', $next_output);
                $output      = $next_output;
                $output_size = 0;
                $o_plugin->SetOutputFile($output);
            }

            #
            # Do we have to interrupt our job ?
            #

            my $stop_all  = $o_params->Get('stop_all');
            my $stop_this = $o_params->Get('stop_this');
            if (-r $stop_all or -r $stop_this or $kld == -2)
            {    # Some .stp file found, or term signal received (qdel)
                if ($i >= $i_end)
                {
                    last;    # The job is finished
                }
                $o_log->Trace("JOB INTERRUPTED BY THE USER\n", '*', '1');
                $o_params->Set('status', 'INTERRUPTED');
                $o_params->Dump($lock_file, NAMESPACE);
                &StopJob($o_params,$o_log);
                exit(0);
            }

        }
    }
    $o_log->Trace("JOB FINISHED - BYE BYE\n", '*', '1');
}

#
# Title: CreatePlugin
# Usage: my $o_plugin = &CreatePlugin($o_params,$o_log)
# function: Using the parameter as a switch, create a 'program' object
# precondition: $o_params and $o_log must exist and be initialized
# Return:   the new object
# Args:     $o_params, $o_log
# Globals:  none
#

sub CreatePlugin
{
    my $o_params = shift;
    my $o_log    = shift;
    
    my $program = $o_params->Get('program');
    my $module  = 'PARALOOP::PLUGIN::' . $program;
    eval("use $module");
    if ($@ ne "")
    {    # Something wrong in eval (may be the object does not work)
        die $@;
    }
    my $o_plugin;
    my $to_eval = "PARALOOP::PLUGIN::$program->New" . '(-params=>$o_params, -log=>$o_log)';
    $o_plugin = eval "$to_eval";
    if ($@ ne "")
    {    # Something wrong in eval (missing object of bad parameter passed
        die $@;
    }

    my $WhatPlugin = $o_plugin->WhatParaloopPlugin();
    print "Found " . $o_plugin->WhatParaloopPlugin() . "\n" unless ($o_params->IsDefined('quiet'));
    return $o_plugin;
}

#
# Title: CreateScheduler
# Usage: my $o_sched = &CreateScheduler($o_params,$o_log)
# function: Using the parameters --local and Scheduler, create a Scheduler object
# precondition: $o_params and $o_log must exist and be initialized
# Return:   the new scheduler object
# Args:     $o_params,$o_log
# Globals:  none
#

sub CreateScheduler
{
    my $o_params = shift;
    my $o_log    = shift;
    my $module    = "";
    my $scheduler = "";

    if ($o_params->IsDefined('Scheduler'))
    {
        $scheduler = $o_params->Get('Scheduler');
    }
    else
    {
        die "The parameter PARALOOP_Scheduler should be defined\n";
    }

    if ($scheduler eq 'System' or $o_params->IsDefined('local'))
    {
        $module = 'PARALOOP::SCHEDULER::System';
    }
    else
    {
        $module = "PARALOOP::SCHEDULER::$scheduler";
    }

    eval("use $module");
    if ($@ ne "")
    {    # Something wrong in eval (may be the object does not work)
        die $@;
    }

    my $cmd;
    my $to_eval = "$module->New" . '(-params=>$o_params, -log=>$o_log)';
    $cmd = eval "$to_eval";
    if ($@ ne "")
    {    # Something wrong in eval (missing object of bad parameter passed
        die $@;
    }

    my $WhatScheduler = $cmd->WhatParaloopScheduler();
    print "Found $WhatScheduler \n" unless ($o_params->IsDefined('quiet'));
    return $cmd;
}

#
# Title: StopJob
# Usage: &StopJob($o_params,$o_log)
# procedure: Log some message and exit.
# Return:   NEVER RETURN - CALL exit
# Args:     $o_params,$o_log
# Globals:  None
#

sub StopJob
{
    my $o_params = shift;
    my $o_log    = shift;
    my $lock_file = $o_params->Get('lock_file');
    $o_log->Trace("INTERRUPTION REQUEST RECEIVED - PLEASE RESTART WITH --restart $lock_file\n", '.', '1');
    exit(0);
}

#
# Title: PrintVar
# Usage: &PrintVar($o_params);
# Procedure: print (to the standard output) the environment
# Return:   none
# Args:     $o_params
# Globals:  %ENV
#

sub PrintVar
{
    my $o_params = shift;
    
    print "\nPARALOOP IMPORTANT ENVIRONMENT VARIABLES\n";
    print "==========================================\n";
    foreach my $v (sort keys %ENV)
    {
        if ($v =~ /^PLP_/ or $v =~ /PBS_/)
        {
            print "$v => $ENV{$v}\n";
        }
    }
    print "\nPARALOOP PARAMETERS\n";
    print "=======================\n";
    my @keys = $o_params->GetKeys('.*');
    foreach my $v (@keys)
    {
        print "$v => " . $o_params->Get($v) . "\n";
    }
}

#
# Title: IncrementName
# Usage: $new_name = &IncrementName($old_name)
# Function: search and increment the last number in the name passed by parameter
#           build and return a new name, incrementing this number
#           If no number found, do nothing
# Return:   the new name
# Args:     the name to increment
# globals:  none
#

sub IncrementName
{
    my $name = shift;
    if ($name =~ /(\d+)([^ \d\/]+)\Z/)
    {
        my $nb = $1;
        $nb++;
        $name =~ s/(\d+)([^ \d\/]+)\Z/$nb$2/;
    }
    return $name;
}

#
# Title: WaitParaloop
# Usage: &WaitParaloop($directory)
# Procedure: Sleep some time (100 s), then tests the existence of $lock
#            Returns only when $lock (a file or a directory) does not exist any more
#            Returns at once if it does not exist at all
# Return:   none
# Args:     $directory = The directory to check
# globals:  none
#

sub WaitParaloop
{
    my $lock = shift;
    sleep 5;
    
    if (-d $lock)
    {
	print "\nNow Waiting for all the subprocesses to finish\n";
	
	#print "ALL THE PROCESSES MAY BE INTERRUPTED WITH THE COMMAND: kill -int $$ \n";
	do
	{
	    if (&IsInterrupted($lock) == 1)
	    {
		print "ALL THE JOBS ARE INTERRUPTED\n";
		print "You can restart the jobs with paraloop.pl --restart $lock --wait\n";
		return;
	    }
	    sleep WAIT_TIME;
	    rmdir $lock;    # Try to remove the directory
	    #print "$lock existe-t-il ?\n";
	} while (-d $lock);
    }
    elsif (-f $lock)
    {
	print "\nNow Waiting for one subprocess to finish (see $lock)\n";
	do
	{
	    if (&IsInterrupted($lock) == 1)
	    {
		print "JOB INTERRUPTED\n";
		print "You can restart the job with paraloop.pl --restart $lock --wait\n";
		return;
	    }
	    sleep WAIT_TIME;
	} while (-f $lock);
   }
}

#
# Title: IsInterrupted
# Usage: $sts = &IsInterrupted()
# Function:  Return 1 if all the jobs were interrupted (if parameter is a directory)
#                     or if the job is interrupted (if parameter is a file)
# Return:    0/1
# Args:      The lock directory
# globals:   none
#

sub IsInterrupted
{
    my $lock = shift;
    if (-d $lock)
    {
        my $all_interrupt = 1;
        foreach my $f (glob("$lock/lock.*"))
        {
            my $o_param = New ParamParser($f);
            $o_param->SelectNameSpace(NAMESPACE);
            my $status = $o_param->Get('status');
            if ($status ne 'INTERRUPTED')
            {
                $all_interrupt = 0;
                last;
            }
        }
        return $all_interrupt;
    }
    elsif (-f $lock)
    {
	my $o_param = New ParamParser($lock);
	$o_param->SelectNameSpace(NAMESPACE);
	my $status = $o_param->Get('status');
	return ($status eq 'INTERRUPTED') ? 1 : 0;
    };
 	
    return 0;
}

#
# Title: ReadCfgFiles
# Usage: &ReadCfgFiles($object)
# Function:  Using the cfg parameter if defined, create a list of configuration files
#            If we are in verbose mode, test the existence of the file and print a message for each file found
#            Update the object passed by reference
#			 Correct some parameters if necessary
# Return:    The list of found configuration files
# Args:      a ParamParser object
# globals:   $0
#

sub ReadCfgFiles
{
    my $o_params = shift;
    my ($dirprg, $nameprg) = $0 =~ /(.+)\/(.+)/;

    # First create an array with the config files
    my @config_files = ();

    # Read the config /etc/paraloop.root.cfg files
    if (exists $ENV{'PARALOOP'})
    {
        push(@config_files,$ENV{'PARALOOP'}.'/etc/paraloop.root.cfg');
    }
    elsif (exists $ENV{'LIPMLIB'})
    {
        push(@config_files, $ENV{'LIPMLIB'} . '/../etc/paraloop.root.cfg');
    }
    else
    {
        push(@config_files, "$dirprg/etc/paraloop.root.cfg", "$dirprg/../etc/paraloop.root.cfg");
    }

    # Read the config files defined by the cfg switch
    if ($o_params->IsDefined('cfg'))
    {
        push(@config_files, split(',', $o_params->Get('cfg')));
    }

    # Read .paralooprc
    push(@config_files, $ENV{'HOME'} . '/.paralooprc');

    # Then, read the etc/paraloop.cfg file

    if (exists $ENV{'PARALOOP'})
    {
        push(@config_files,$ENV{'PARALOOP'}.'/etc/paraloop.cfg');
    }
    elsif (exists $ENV{'LIPMLIB'})
    {
        push(@config_files, $ENV{'LIPMLIB'} . '/../etc/paraloop.cfg');
    }
    else
    {
        push(@config_files, "$dirprg/etc/paraloop.cfg", "$dirprg/../etc/paraloop.cfg");
    }

    # Build the list of really found configuration files
    # If verbose, display which file exists and which file does not exist
    my @found_config_files;
    foreach my $f (@config_files)
    {
        if ($o_params->IsDefined('verbose'))
        {

            print "Configuration file $f";
            (-r $f) ? print " OK\n" : print " DOES NOT EXIST !\n";
        }
        push(@found_config_files, $f) if (-r $f);
    }

    # Update the ParamParser object from those files
    $o_params->SetBehaviour("ignore_space");
    $o_params->SelectNameSpace();    # Cancel the namespace before reading the conf files

    foreach my $f (@found_config_files)
    {
        $o_params->Update($f, 'A');
    }
    $o_params->SelectNameSpace(NAMESPACE);

    # Corrrect some parameters

    if ($o_params->IsDefined('no_local_mode'))    # May be set by the administrator
    {
        if ($o_params->IsDefined('local') and $o_params->Get('local') == 1)
        {
            print "SORRY, the local mode is disabled on this site - Terminating\n";
            exit(0);
        }
    }
    if ($o_params->IsDefined('local') and $o_params->Get('local') == 1)
    {
        $o_params->Delete('queue') if ($o_params->IsDefined('local'));    # local has the precedence
        $o_params->Delete('fair_time_limit');                             # No fair_time in local mode
    }

    return \@found_config_files;
}

#
# Title: Usage,Switches,Parameters,Plugins
# Usage: &Usage(),Switches(),Parameters(),Plugins()
# Function: print some documentation...
# Return:   none
# Args:     $plugin_debug: print more things if 1
# globals:  none
#

sub Usage
{
    print "The complete documentation: paraloop.pl [--switches|--parameters|--plugins|--schedulers|--help]\n";
    print "If you are in a hurry, you could try:\n\n";
    print "       paraloop.pl --program=SomePlugin [--db=database] --input=infile --output=outputfile\n";
    print "       paraloop.pl --help\n";
}

sub Switches
{
    print "\nThe switches of general use \n";
    print "--cfg=f1,f2,f3   \tThe list of configuration files\n";
    print "--verbose       \tsome verbosity added\n";
    print "--quiet         \tless messages\n";
    print "--plugin_debug  \tThis special mode is only useful for debugging the plugins\n";
    print "--switches      \tThis screen\n";
    print "--parameters    \tThe parameters you can put in configuration files\n";
    print "--plugins       \tThe list of installed plugins\n";
    print "--plugins --debug=D \tMay be useful to debug a new plugin\n";
    print "--help          \tA very quick help screen\n";
    print "\nThe switches related to PBS \n";
    print "--queue=string  \tThe queue the jobs will be submitted to\n";
    print "--account=string\tAccount name, passed to qsub\n";
    print "--local         \tExecute the jobs on the local machine (But see also no_local_mode)\n";
    print "\nThe switches related to the program running through the loop\n";
    print "--program       \tThe name of a Perl object to run inside the loop\n";
    print "                \tExamples: Dummy, Shell (try perldoc -I \$PARALOOP PARALOOP::Shell)\n";
    print "--input         \tThe input file/database/directory name\n";
    print "--output=OUT    \tThe output prefix-filename. Real name will be:\n";
    print "                \tOUT.X.Y.txt (output file) and OUT.X.Y.log (log file)\n";
    print "--db=string     \tOne more paramter used by some objects (Blastp, Blastpgp,..)\n";
    print "\nSpecifying start - end - step switches\n";
    print "--start=integer \tRecord number (1st = 0) to begin the loop with, including this number\n";
    print "--end=integer   \tRecord number to end the loop with, including this number\n";
    print "--step=integer  \tStep used by the loop. Default 1. May be negative if the object agrees.\n";
    print "                \tIf --step is negative, start must be higher than end\n";
    print
      "\nSpecifying slices: the input file is divided into equal-sized slices, each instance works on a specific slice.\n";
    print "The Start and End parameters are computed from the slice_offset, slice_size and slice_nb parameters.\n";
    print "The step parameter is ALWAYS 1\n";
    print "--slice_offset=integer\tThe offset to take into account for computing the start\n";
    print "--slice_size=integer\tThe size of each slice (number of records)\n";
    print "--slice_nb=interger\tThe slice number to use\n";
    print "\nSpecifying cpus: the number of instances is automagically computed\n";
    print "--monoprocessor\tRun the job ONLY in one processor\n";
    print "--ncpus=integer\tThe number of paraloop instances to start, generally less than the number of cpus\n";
    print "--start=integer\tThe global start record\n";
    print "--end=integer\tThe global end record\n";
    print "--interleaved\tIf specified, The records are distributed on a round-robin basis, NOT by slices\n";
    print "--wait\tWait until every child process returns\n";
    print "\nInterrupting the jobs\n";
    print "1st method\tqdel -W 2 job_Id\n";
    print "2nd method\ttouch lock/paraloop.X.stp to stop the instance which started at record X\n";
    print "3rd method\ttouch lock/paraloop.stp to stop every instance\n";

    print "\nRestarting after an interruption\n";
    print "--restart=lock\tRestart every job whose lock file lives in this directory\n";
    print "--restart=lock/lock.x.lck\tRestart Only the job whose lock file is specified\n";

    print "\nWaiting for the jobs to finish\n";
    print "The switch --waitonly can be specified ALONE - paraloop makes NOTHING, just waiting\n";
    print "--waitonly=lock\tWait until ALL the jobs are finished OR interrupted\n";
    print "--waitonly=lock/lock.x.lck\tWait until this job is finished or interrupted\n";
   
    print "\n";
    exit(0);
}

sub Parameters
{
    my $plugin_debug = shift;
    my ($dirprg, $nameprg) = $0 =~ /(.+)\/(.+)/;

    print "\nThe PARALOOP parameters\n\n";
    print "The parameter name is ALWAYS preceded by PARALOOP_\n";
    print "\nNearly all switches can be specified as a parameter, in a configuration file\n";
    print "Example: --queue workq can be specified as PARALOOP_queue=workq\n";
    print "\nThe parameters files may live in several standard locations, and you may specify additional files\n";
    print "\nEx.: paraloop.pl ... --cfg f1.cfg,f2.cfg,f3.cfg gives following priorities (highest to lowest):\n";
    print "\t1. \tThe switch on the command line\n";
    print "\t2. \t\$PARALOOP/etc/paraloop.root.cfg\n";
    print "\t3. \tf1.cfg\n";
    print "\t4. \tf2.cfg\n";
    print "\t5. \tf3.cfg\n";
    print "\t6. \t\$HOME/.paralooprc\n";
    print "\t7. \t\$PARALOOP/paraloop.cfg\n";
    print "\nParameters for the PARALOOP script [see also $0 --switches()]\n=================\n";
    print "The parameters described here cannot be specified on the command line\n\n";
    print "PARALOOP_log_level\t0\t-> do not log anything\n";
    print "                    \t01\t-> log more things\n";
    print "                    \t012\t-> log a lot of things\n\n";
    print "PARALOOP_no_local_mode\tWhen specified, the switch --local cannot be used.\n";
    print "PARALOOP_fair_time_limit\tWhen the time elapsed since the beginning of the loop reachs this value (in s),\n";
    print
      "                     \tthe loop is interrupted and another qsub is submitted. However, this NOT managed as a timeout\n";
    print "                     \tso that your running job should not be inexpectedly interrupted\n";
    print "                     \tThus, it is fair specifying this parameter\n";
    print "PARALOOP_max_file_size\tThe output file is NEVER bigger thant this size in Mb (default 1000)\n";
    print "PARALOOP_error_directory\tThe directory for the storage of output/error files\n";
    print "PARALOOP_lock_directory\tThe lock directory\n";

    my $lib = '';
    foreach my $d (@INC)
    {
        $lib = $d . '/' . 'PARALOOP/PLUGIN';
        if (-d $lib)
        {
            foreach my $f (glob("$lib/*.pm"))
            {    # For file
                chomp($f);
                if ($f =~ /$lib\/(.+)\.pm\Z/)
                {
                    my $module = 'PARALOOP::PLUGIN::' . $1;
                    eval("use $module");
                    print $@ if ($plugin_debug != 0);
                    next if ($@ ne "");
                    my $msg;
                    my $plugin = $module . '::Parameters' . '()';
                    $msg = eval($plugin);
                    print $@ if ($plugin_debug != 0);
                    next if ($@ ne "");
                    print "$msg\n";
                }
            }
        }

        $lib = $d . '/' . 'PARALOOP/SCHEDULER';
        if (-d $lib)
        {
            foreach my $f (glob("$lib/*.pm"))
            {    # For file
                chomp($f);
                if ($f =~ /$lib\/(.+)\.pm\Z/)
                {
                    my $module = 'PARALOOP::SCHEDULER::' . $1;
                    eval("use $module");
                    print $@ if ($plugin_debug != 0);
                    next if ($@ ne "");
                    my $msg;
                    my $plugin = $module . '::Parameters' . '()';
                    $msg = eval($plugin);
                    print $@ if ($plugin_debug != 0);
                    next if ($@ ne "");
                    print "$msg\n";
                }
            }
        }
    }
    exit(0);
}

sub Plugins
{
    my $plugin_debug = shift;
    print "\nCurrently installed Plugins\n\n";
    foreach my $d (@INC)
    {
        my $lib = $d . '/' . 'PARALOOP/PLUGIN';
        if (-d $lib)
        {
            foreach my $f (glob("$lib/*.pm"))
            {    # For file
                chomp($f);
                if ($f =~ /$lib\/(.+)\.pm\Z/)
                {
                    my $module = 'PARALOOP::PLUGIN::' . $1;
                    eval("use $module");
                    print $@ if ($plugin_debug != 0);
                    next if ($@ ne "");
                    my $msg;
                    my $plugin = $module . '::WhatParaloopPlugin' . '()';
                    $msg = eval($plugin);
                    print $@ if ($plugin_debug != 0);
                    next if ($@ ne "");
                    print "$msg\n";
                }
            }
        }
    }
    exit(0);
}

sub Schedulers
{
    my $plugin_debug = shift;
    print "\nCurrently installed Schedulers\n\n";
    foreach my $d (@INC)
    {
        my $lib = $d . '/' . 'PARALOOP/SCHEDULER/';
        if (-d $lib)
        {
            foreach my $f (glob("$lib/*.pm"))
            {    # For file
                chomp($f);
                if ($f =~ /$lib\/(.+)\.pm\Z/)
                {
                    my $module = 'PARALOOP::SCHEDULER::' . $1;
                    eval("use $module");
                    print $@ if ($plugin_debug != 0);
                    next if ($@ ne "");
                    my $msg;
                    my $plugin = $module . '::WhatParaloopScheduler' . '()';
                    $msg = eval($plugin);
                    print $@ if ($plugin_debug != 0);
                    next if ($@ ne "");
                    print "$msg\n";
                }
            }
        }
    }
    exit(0);
}

