
# $Id: Plugin.pm 70 2005-06-06 09:26:58Z manu $

package PARALOOP::PLUGIN::Plugin;

=pod

=head1 Copyright Notice

 Copyright I.N.R.A. - C.N.R.S. 

 emmanuel.courcelle@toulouse.inra.fr
 jerome.gouzy@toulouse.inra.fr

 This software is a computer program whose purpose is to help you run 
 your scripts on parallel computers.

 This software is governed by the CeCILL  license under French law and
 abiding by the rules of distribution of free software.  You can  use, 
 modify and/ or redistribute the software under the terms of the CeCILL
 license as circulated by CEA, CNRS and INRIA at the following URL
 http://www.cecill.info" 

 As a counterpart to the access to the source code and  rights to copy,
 modify and redistribute granted by the license, users are provided only
 with a limited warranty  and the software's author,  the holder of the
 economic rights,  and the successive licensors  have only  limited
 liability. 

 In this respect, the user's attention is drawn to the risks associated
 with loading,  using,  modifying and/or developing or reproducing the
 software by the user in light of its specific status of free software,
 that may mean  that it is complicated to manipulate,  and  that  also
 therefore means  that it is reserved for developers  and  experienced
 professionals having in-depth computer knowledge. Users are therefore
 encouraged to load and test the software's suitability as regards their
 requirements in conditions enabling the security of their systems and/or 
 data to be ensured and,  more generally, to use and operate it in the 
 same conditions as regards security. 

 The fact that you are presently reading this means that you have had
 knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

PARALOOP::PLUGIN::Plugin - Abstract class for the Plugin objects

=head1 Derives from...

__Initializable 

=head1 DESCRIPTION

The objects deriving from Plugin implement 5 methods, requested by paraloop. Plugin calls dies
for some of those, thus declaring abstract methods.

=head1 PUBLIC METHODS


=head2 I/O METHODS (all abstract)

 NextRecord
    $obj->NextRecord();

 SkipRecords
    $obj->SkipRecords(10)

 Tell
    my $record_number = $obj->Tell();

 TellLength
    my $nb_of_records = $obj->TellLength();

=head2 OTHER METHODS

 WhatParaloopPlugin
    my $msg=$obj->WhatParaloopPlugin();
    ABSTRACT

 Parameters
    my $msg=$obj->Parameters();
    ABSTRACT
    
 New
    my $obj = PARALOOP::Dummy->New(-params => $params, -log => $log);
    ABSTRACT (cf __Init)

 Exec
    my ($incr_size,$error,$killed) = $obj->Exec();
    ABSTRACT

 SetOutputFile
    $obj->SetOutputFile('filename');
    
 __JobId
    my $job_id = $obj->__JobId($o_param)


=head1 METHODS AND SUBROUTINES - DETAILED DESCRIPTION

=cut

use strict;
use warnings;
use Cwd;
use File::Basename;
use Carp;

use __Initializable;
use Logger;
use Runner;

@PARALOOP::PLUGIN::Plugin::ISA = qw( __Initializable );

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 70 $ =~ /\d+/g); $r[0]};

}

=pod

=head2 Function WhatParaloopPlugin

 Usage:    my $msg = PARALOOP::Dummy::WhatParaloopPlugin()
        or my $msg = $obj->WhatParaloopPlugin()
 Function: Return some information (2 lines max) on the plugin
           An object is NOT considered as a plugin if this function is not implemented
 Arg:      none ($self is ignored)
 Return:   A message describing the plugin (1 or 2 lines)

=cut

sub WhatParaloopPlugin {die "\n";}

=pod

=head2 Function Parameters

 Usage:    my $msg = PARALOOP::Dummy::Parameters()
        or my $msg = $obj->Parameters()
 Function: Return the list of allowed parameters for this plugin
           If your plugin defines some parameters, it is good pratice adding this function,
           to display those parameters when the --parameters switch of paraloop.pl is used.
           However, this is not mandatory
 Arg:      none ($self is ignored)
 Return:   A message describing the plugin (1 or 2 lines)

=cut

sub Parameters {die "\n";}

=pod

=head2 procedure __Init

 Usage:  CALLED BY _Initializable::New
         my obj = PARALOOP::Dummy->New(-params => $params, -log => $log);
 Access: private (use New instead)
 procedure: Init the object
 Return:   nothing
 Args:     -params = an object of class ParamParser (already initialized)
           -log    = an object of class Log (already initialized)
 globals:  none

=cut

sub __Init {die "\n";}

=pod

=head2 Procedure SetOutputFile

 Usage:     &$opt->SetOutputFile('filename');
 Procedure: change the name of the output file
 Arg :      the output file name

=cut

sub SetOutputFile
{
    my $self       = shift;
    my $outputfile = shift;
    $self->{'__outputfile'} = $outputfile;
}

=head2 Function Exec

 Usage:    my ($incr_size, $error,$killed) = &$opt->Exec($timeout);
 Function: Copy the input record to the output
           Sleep a few seconds
 Arg:      $timeout   = The number of seconds after which a timeout is sent
 Return:   $incr_size = The number of bytes added to the output file
           $error     = An error code. 
		                If != 0, paraloop WILL BE TERMINATED - Generally marks the end of file
           $killed    = A kill code
		                0: everything OK 
                       -1: killed by timeout
                        -2: killed by TERM signal (kill, qdel, ...)

=cut

sub Exec {die "\n";}

=pod

=head2 function NextRecord

 Usage:    my $line = $obj->NextRecord()
 function: read and return next record (a string)
           update private member __records_counter
 Return:   a string
 Args:     none

=cut

sub NextRecord {die "\n";}

=pod

=head2 function Tell

 Usage:    my $cnt = $obj->Tell()
 function: return records_counter private field
 Return:   The counter of records (first=0)
 Args:     none

=cut

sub Tell {die "\n";}

=pod

=head2 function TellLength

 Usage:    my $cnt = $obj->TellLength()
 function: Open the file again, call next_seq() until the end the file is reached, 
           counting the number of records in the file.
 Return:   The total number of records
 Args:     none

=cut

sub TellLength {die "\n";}

=pod

=head2 procedure SkipRecords

 Usage:    $obj->SkipRecords($n)
 procedure:Position the pointer, skipping $n records
 Args:     The number of records to skip (MUST BE > 0)

=cut

sub SkipRecords {die "\n";}

=head2 function _JobId

 Usage:    my $job_id = $obj->__JobId($o_param)
 function: Return the job_id, as found in the ParamParser object.
           If 0, return a pseudo job_id, just good enough to create a temporary directory of file
 Args:     A ref to a ParamParsers object
 Access:   protected

=cut

sub _JobId
{
    my $self    = shift;
    my $o_param = shift;
    my $job_id  = $o_param->IsDefined('job_id') ? $o_param->Get('job_id') : int(rand 1000000000);
    return $job_id;
}

1;

