
# $Id: Rsystem.pm 70 2005-06-06 09:26:58Z manu $

package PARALOOP::SCHEDULER::Rsystem;

=pod

=head1 Copyright Notice

 Copyright I.N.R.A. - C.N.R.S. 

 emmanuel.courcelle@toulouse.inra.fr
 jerome.gouzy@toulouse.inra.fr

 This software is a computer program whose purpose is to help you run 
 your scripts on parallel computers.

 This software is governed by the CeCILL  license under French law and
 abiding by the rules of distribution of free software.  You can  use, 
 modify and/ or redistribute the software under the terms of the CeCILL
 license as circulated by CEA, CNRS and INRIA at the following URL
 http://www.cecill.info" 

 As a counterpart to the access to the source code and  rights to copy,
 modify and redistribute granted by the license, users are provided only
 with a limited warranty  and the software's author,  the holder of the
 economic rights,  and the successive licensors  have only  limited
 liability. 

 In this respect, the user's attention is drawn to the risks associated
 with loading,  using,  modifying and/or developing or reproducing the
 software by the user in light of its specific status of free software,
 that may mean  that it is complicated to manipulate,  and  that  also
 therefore means  that it is reserved for developers  and  experienced
 professionals having in-depth computer knowledge. Users are therefore
 encouraged to load and test the software's suitability as regards their
 requirements in conditions enabling the security of their systems and/or 
 data to be ensured and,  more generally, to use and operate it in the 
 same conditions as regards security. 

 The fact that you are presently reading this means that you have had
 knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

PARALOOP::SCHEDULER::Rsystem - Scheduler object for use with clusters without any batch system

=head1 Derives from...

PARALOOP::SCHEDULER::Scheduler

=head1 DESCRIPTION

This scheduler is used for a cluster, when there is no batch system installed: you have to distribute the jobs
through remote commands (rsh or ssh).

=head1 CONFIGURATION FILE PARAMETERS

 PARALOOP_Scheduler        
     Rsystem
 PARALOOP_error_directory  
     The name of the error directory
 PARALOOP_Rsystem_nodes    
     The list of nodes constituting the cluster
 PARALOOP_Rsystem_rsh      
     The remote shell program (defaults to rsh)
 PARALOOP_Rsystem_tmp      
     The name of the temporary directory - SHOULD BE LOCAL TO EACH NODE

=head1 PUBLIC METHODS

 WhatParaloopScheduler
    my $msg=$obj->WhatParaloopScheduler()

 Parameters
    my $msg=$obj->Parameters()

 New
    my $obj = PARALOOP::Dummy->New(-params => $params, -log => $log);

 Submit
    my $job_id = $obj->Submit();

 RemTime
    my $rem_time = $obj->RemTime(); 

 JobId
    my $job_id = $obj->JobId();

=head1 METHODS AND SUBROUTINES - DETAILED DESCRIPTION

=cut

use strict;
use warnings;
use English;
use File::Basename;
use Carp;

# The NAMESPACE constant is also set in paraloop
use constant NAMESPACE => 'PARALOOP_';

use Runner;
use PARALOOP::SCHEDULER::Scheduler;

@PARALOOP::SCHEDULER::Rsystem::ISA = qw( PARALOOP::SCHEDULER::Scheduler );

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 70 $ =~ /\d+/g); $r[0]};

}

=pod

=head2 Function WhatParaloopScheduler

 Usage:    my $msg = PARALOOP::SCHEDULER::System::WhatParaloopScheduler()
        or my $msg = $obj->WhatParaloopScheduler()
 Function: Return some information (2 lines max) on the scheduler
           An object is NOT considered as a scheduler if this function is not implemented
 Arg:      none ($self is ignored)
 Return:   A message describing the plugin (1 or 2 lines)

=cut

sub WhatParaloopScheduler
{
    my $rvl = "";
    $rvl =
      "PARALOOP::SCHEDULER::Rsystem v$PARALOOP::SCHEDULER::Rsystem::VERSION - Deriving from Scheduler v$PARALOOP::SCHEDULER::Scheduler::VERSION - Using Runner v$Runner::VERSION\n";
    $rvl .= "This scheduler object distributes the job through an rsh-types program\n";
    return $rvl;
}

=pod

=head2 Function Parameters

 Usage:    my $msg = PARALOOP::SCHEDULER::System::Parameters()
        or my $msg = $obj->Parameters()
 Function: Return the list of allowed parameters for this scheduler
           If your scheduler defines some parameters, it is good pratice adding this function,
		   to display those parameters when the --parameters switch of paraloop.pl is used.
		   However, this is not mandatory
 Arg:      none ($self is ignored)
 Return:   A message describing the scheduler parameters (1 line / parameter)

=cut

sub Parameters
{
    my $rvl = "";
    $rvl = "\nParameters for the Rsystem scheduler\n=================\n";
    $rvl .= "PARALOOP_Scheduler\tMust be set to Rsystem for using this scheduler\n";
    $rvl .= "PARALOOP_Rsystem_nodes\tThe list of nodes belonging to the cluster\n";
    $rvl .= "PARALOOP_Rsystem_rsh\tThe remote command (defaults to rsh)\n";
    $rvl .= "PARALOOP_Rsystem_tmp\tThe temporary directory on each node (defaults to /tmp)\n";
    return $rvl;
}

=head2 procedure __Init

 Usage:  CALLED BY _Initializable::New
         my $obj = PARALOOP::Dummy->New(-params => $params, -log => $log);
 Access: private (use New instead)
 procedure: Init the object
 Return:   nothing
 Args:     -params = an object of class ParamParser (already initialized)
           -log    = an object of class Logger (already initialized)
 globals:  none

=cut

sub __Init
{
    my $self = shift;
    $self->SUPER::__Init(@_);
    my $o_param = $self->{'__params'};
    my $o_log   = $self->{'__log'};

    #
    # If exists, read the Rsystem_rscript_name parameter (hidden)
    #

    if ($o_param->IsDefined('Rsystem_rscript_name'))
    {
        $self->{'__rscript_name'} = $o_param->Get('Rsystem_rscript_name');
    }

    #
    # If we have a subcount, we redirect the Io and we leave
    #

    if ($o_param->IsDefined('subcount') and ($o_param->Get('subcount')) >= 1)
    {
        $self->IoRedirect();
        return;
    }

    #
    # Read the rsh and cluster_nodes parameters
    # die if no cluster_nodes specified
    #

    if ($o_param->IsDefined('Rsystem_nodes'))
    {
        my @rsystem_nodes = split(',', $o_param->Get('Rsystem_nodes'));
        $self->{'__rsystem_nodes'} = \@rsystem_nodes;
    }
    else
    {
        confess('The Rsystem_nodes parameter is NOT defined');
    }

    if ($o_param->IsDefined('Rsystem_rsh'))
    {
        $self->{'__rsh_bin'} = $o_param->Get('Rsystem_rsh');
    }
    else
    {
        $self->{'__rsh_bin'} = "rsh";    # Hope rsh is in the path
    }

    #
    # Read the tmp and cluster_nodes parameter
    #

    if ($o_param->IsDefined('Rsystem_tmp'))
    {
        $self->{'__tmp'} = $o_param->Get('Rsystem_tmp');
    }
    else
    {
        $self->{'__tmp'} = "/tmp";    # /tmp should be local to each node
    }

}

=pod

=head2 Procedure DESTROY

 Usage     Automatically called when the object is destroyed
 Function  Remove the temporary file, is it exists

=cut

sub DESTROY
{
    my $self = shift;
    if (exists($self->{'__rscript_name'}))
    {
        my $rscript_name = $self->{'__rscript_name'};
        unlink $rscript_name;
    }
}

=pod

=head2 Function Submit

 Usage:     &$obj->Submit()
 Function:  Run paraloop on another node
 Arg :      none
 Return:    The job Id

=cut

sub Submit
{
    my $self    = shift;
    my $o_param = $self->{'__params'};
    my $o_log   = $self->{'__log'};

    my $subcount = $self->_SubCount();
    if ($subcount > 1)
    {
        $o_log->Trace("ERROR - THIS SCHEDULER CANNOT RESUBMIT ITSEF\n", '*', '1');
        confess(
            'With this scheduler, paraloop should resubmit itself ONLY 1 time. May be PARALOOP_fair_time_limit is specified ?'
            );
    }

    my $node = $self->__NextNode();
    my $rcmd = $self->{'__rsh_bin'};
    my $tmp  = $self->{'__tmp'};

    # Build a little sh-script on the node

    my $rscript_name = $tmp . '/paraloop.' . $$ . '.' . time() . '.sh';
    $rcmd .= " $node cat '>' $rscript_name";

    open(RCMD, "| $rcmd") or confess("Cannot call $rcmd");
    print(RCMD "#! /bin/sh\n");
    print(RCMD $self->__PrintEnv());

    # Pass to the paraloop instance on the remote node a "hidden" parameter: this is the name of the temporary script
    # Paraloop will be able to remove the file before dying

    print(RCMD "export PARALOOP_Rsystem_rscript_name=$rscript_name\n");
    print(RCMD "$0\n");
    close(RCMD);

    #
    # Make the file executable
    #

    $rcmd = $self->{'__rsh_bin'};
    $rcmd .= " $node chmod u=rx $rscript_name";
    `$rcmd`;

    #
    # Execute the script
    #

    $rcmd = $self->{'__rsh_bin'};
    $rcmd .= " $node $rscript_name";
    print "calling $rcmd\n" if ($o_param->IsDefined('verbose'));

    my $job_id = &RunExtNoWait(-cmd => $rcmd);
    $self->_KeepJobId($job_id);
    $o_log->Trace("$0\n", '', '2');
    my $master = `hostname`;
    chomp($master);
    return $job_id . '.' . $master . '#' . $node;
}

=pod

=head2 Function JobId

 Usage:     my $job_id = $obj->JobId()
 Function:  Ask the system for the job id
            The variable $$ is returned
 Arg :      none
 Return : the job Id

=cut

sub JobId
{
    my $j = $$ . '.' . `hostname`;
    chomp($j);
    return $j;
}

=pod

=head2 Function __NextNode

 Usage:     &$obj->NextNode()
 Function:  return the name of the next "available" node
 Arg :      none

=cut

sub __NextNode
{
    my $self = shift;
    my $tmp  = $self->{'__tmp'};

    my $ra_nodes = $self->{'__rsystem_nodes'};
    my $node     = '';

    my $node_file = $tmp . "/paraloop_rsystem_node_$UID";
    my $lock_file = $tmp . "/paraloop_rsystem_lock_$UID";

    open(LOCK, "> $lock_file") or confess("Cannot open $lock_file for writing");
    flock(LOCK, 2) or confess("Cannot lock $lock_file");    # Request an exclusive lock on this file

    my $old_node;
    if (open(NODER, "$node_file"))
    {
        $old_node = <NODER>;
        close(NODER);
    }

    if (defined($old_node))
    {
        chomp($old_node);
    }
    else
    {
        $old_node = $$ra_nodes[0];
    }

    $node = $old_node;
    my $cnt = @$ra_nodes;
    while (1)
    {
        $node = $self->__IncrNode($ra_nodes, $node);
        last if ($self->__IsAlive($node));
        $cnt--;
        last if ($cnt <= 0);
    }

    if ($cnt <= 0)
    {
        confess("ERROR - ALL NODES ARE DEAD");
    }

    open(NODEW, "> $node_file") or confess("Cannot open $node_file for writing");
    print NODEW "$node\n";
    close(NODEW);

    flock(LOCK, 8) or confess("Cannot release $lock_file");    # Release the lock

    return $node;
}

=pod

=head2 Function __IncrNode

 Usage:     $node = &$obj->__IncrNode(\@nodes,'node3')
 Function:  return the name of the next node on the list
 Arg :      \@nodes Ref to an array of nodes
            $position The current node in the list
 Return:    The next node on the list, or $$nodes[0] if $position is not on the list	    

=cut

sub __IncrNode
{
    my $self     = shift;
    my $ra_nodes = shift;
    my $position = shift;

    my $i = 0;
    for ($i = 0; $i <= $#$ra_nodes; $i++)
    {
        last if ($$ra_nodes[$i] eq $position);
    }
    $i = ($i >= $#$ra_nodes) ? 0 : ++$i;
    return $$ra_nodes[$i];
}

=pod

=head2 Function __IsAlive

 Usage:     if ($obj->__IsAlive($node)) ...
 Function:  return 1 if the node is alive
            We check the node with asking him its name
 Arg :      The name of the node to check
 Return:    1 if alive, 0 if dead

=cut

sub __IsAlive
{
    my $self    = shift;
    my $node    = shift;
    my $o_param = $self->{'__params'};

    my $rcmd = $self->{'__rsh_bin'};
    $rcmd .= " $node hostname";
    print "calling $rcmd\n" if ($o_param->IsDefined('verbose'));

    # The host has 2 seconds to return its name
    # If it cannot return, because it is dead... or too busy, we consider it is dead

    my ($status, $killed) = &RunExt(-cmd => $rcmd, -timeout => 2, -stdout => ">/dev/null", -stderr => ">/dev/null");

    #    my ($status, $killed) = &RunExt(-cmd=>$rcmd);
    if ($status == 0 and $killed == 0)
    {
        return 1;    # No problemo, we are alive
    }
    else
    {
        return 0;    # Hum... Very busy, somewhat ill, or completely dead
    }
}

=pod

=head2 Function __PrintEnv

 Usage:     my $env = $obj->PrintEnv()
 Function:  For each environment variable starting with NAMESPACE,
            append to $env export ENV='$ENV{ENV}'
 Arg :      none
 Return :   $env

=cut

sub __PrintEnv
{
    my $env    = '';
    my $prefix = NAMESPACE;
    foreach my $v (keys %ENV)
    {
        if ($v =~ /\A$prefix/)
        {
            $env .= "export $v='" . $ENV{$v} . "'\n";
        }
    }
    return $env;
}

=pod

=head2 Function IsLaunched

 Usage:    If (&$obj->IsLaunched($job_id)==1)...
 Function: Check the presence of the job in the system
 Arg:      none
 Return:   0 ==> The job is not known by the system
           1 ==> The job is known (may be in different states: Run, Queue, etc)

=cut

sub IsLaunched
{
    my $self    = shift;
    my $o_param = $self->{'__params'};

    my $job_id = shift;
    my $ps_cmd = '';
    if ($job_id =~ /(\d+)\.([a-z0-9_]+)/)    # Ex 3456.node1
    {
        my $node = $2;
        $job_id = $1;
        my $rsh = $self->{'__rsh_bin'};
        $ps_cmd = "$rsh $node \"ps ax | egrep '^ *" . $job_id . "'\" | wc -l";
        print "calling $ps_cmd\n" if ($o_param->IsDefined('verbose'));
    }
    else
    {
        $ps_cmd = "ps ax | egrep '^ *" . $job_id . "' | wc -l";
        print "calling $ps_cmd\n" if ($o_param->IsDefined('verbose'));
    }

    my $rvl = `$ps_cmd`;
    chomp($rvl);
    return $rvl;
}

=pod

=head2 Procedure IoRedirect

 Usage:     &$obj->IoRedirect()
 Procedure: Redirect stdout and stderr to some file ($job_id.EO)
            We DO NOT close STDIN, because it will be impossible later to redirect STDOUT
	        (I think there is due to a bug in perl...)
 Arg :      none
 Return:    none

=cut

sub IoRedirect
{
    my $self            = shift;
    my $o_param         = $self->{'__params'};
    my $error_directory = $o_param->Get('error_directory');
    my $job_id          = $self->JobId();
    my $stdoe           = $error_directory . '/' . $job_id . '.EO';

    #    close(STDIN);
    close(STDOUT);
    close(STDERR);
    open(STDOUT, "> $stdoe")  or Confess("Cannot open output file");
    open(STDERR, ">> $stdoe") or Confess("Cannot open error file");
    warn("now STDOUT and STDERR redirected\n");
}

1;

